#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul 19 17:00:20 2024

@author: thahn
"""

# =============================================================================
# Functions to filter output from CoCoMET into something clearner
# =============================================================================

import numpy as np
from tqdm import tqdm

from .analysis import Analysis_Object


# TODO: Does this file do anything? I don't see it implemented anywhere
def filter_cells(
    analysis_object: dict | Analysis_Object,
    **args: dict,
) -> Analysis_Object:
    """


    Parameters
    ----------
    analysis_object : dict
        A  CoCoMET-US standard analysis object containing at least US_features, US_tracks, US_segmentation_2d, tracking_xarray, and segmentation_xarray.
    **args : dict
        Throw away inputs

    Raises
    ------
    Exceptioon
        Exception if missing segmentation data from the analysis object.

    Returns
    -------
    analysis_object_filtered : Analysis_Object
        A  CoCoMET-US standard analysis object containing US_features, US_tracks, US_segmentation_2d, tracking_xarray, and segmentation_xarray.
    """

    # Remove cells which exist at frame 0 and cells which exist at the very last frame
    # Remove cells which get too close to the boundary edge
    # Remove cells which travel outside of the domain defined by an input shapely file

    if type(analysis_object) != dict:
        analysis_object = analysis_object.return_analysis_dictionary()

    frame_groups = analysis_object["US_linking"].groupby("frame")

    if analysis_object["US_segmentation_3d"] is not None:
        cell_seg = analysis_object["US_segmentation_3d"].Cell_Segmentation
        dim = "3d"

    elif ["US_segmentation_2d"] is not None:
        cell_seg = analysis_object["US_segmentation_2d"].Cell_Segmentation
        dim = "2d"

    else:
        raise Exception("!=====Missing Segmentation Input=====!")

    edge_cells_to_delete = []

    # Loop over frames
    for frameg in frame_groups:
        frame = frameg[0]

        # Remove all cells in the first or last frames
        if frame == 0 or (frame == frame_groups.ngroups - 1):
            #     # add all cells in these frames to the delete list
            for cellg in frameg[1].groupby("cell_id"):
                edge_cells_to_delete.append(cellg[0])

        # Go through the sides of the 2D or 3D space and remove
        else:
            # go through and check the top, bottom, left, right, forward, back faces

            bottom_features = np.unique(cell_seg[frame, 0, :, :])[
                1:
            ]  # each will have -1 for a lack of a feature
            top_features = np.unique(cell_seg[frame, -1, :, :])[1:]
            forward_features = np.unique(cell_seg[frame, :, 0, :])[1:]
            backward_features = np.unique(cell_seg[frame, :, -1, :])[1:]

            edge_features = np.unique(
                np.concatenate(
                    (bottom_features, top_features, forward_features, backward_features)
                )
            )
            if dim == "3d":
                left_features = np.unique(cell_seg[frame, :, :, 0])[1:]
                right_features = np.unique(cell_seg[frame, :, :, -1])[1:]

                edge_features = np.unique(
                    np.concatenate((edge_features, left_features, right_features))
                )

            edge_cells_to_delete = edge_cells_to_delete + edge_features.tolist()

    # Make sure the each edge is only listed once
    edge_cells_to_delete = np.unique(edge_cells_to_delete)

    # Dataframes to remove features from
    US_features = analysis_object["US_features"]
    US_linking = analysis_object["US_linking"]

    # Xarrays to remove features from converted to numpy arrays
    # create numpy arrays so hopefully computation will be faster, then revert back to DataArrays when everything is finished

    US_segmentation_2d = analysis_object["US_segmentation_2d"]
    US_segmentation_2d_cell_numpy = US_segmentation_2d.Cell_Segmentation.values
    US_segmentation_2d_feature_numpy = US_segmentation_2d.Feature_Segmentation.values
    if dim == "3d":
        US_segmentation_3d = analysis_object["US_segmentation_3d"]
        US_segmentation_3d_cell_numpy = US_segmentation_3d.Cell_Segmentation.values
        US_segmentation_3d_feature_numpy = (
            US_segmentation_3d.Feature_Segmentation.values
        )
    else:
        US_segmentation_3d = None

    # get dimensions for re-conversion of numpy arrays to DataArrays
    dims2d = US_segmentation_2d.dims
    if dim == "3d":
        dims3d = US_segmentation_3d.dims

    for cell in tqdm(
        edge_cells_to_delete,
        desc="=====Deleting Edge Cells=====",
        total=len(edge_cells_to_delete),
    ):
        cell_indices_2d = where_replacement(US_segmentation_2d_cell_numpy == cell)
        features = -np.ones_like(US_segmentation_2d_cell_numpy)

        for t, y, x in cell_indices_2d:
            features[t, y, x] = US_segmentation_2d_feature_numpy[t, y, x]

            US_segmentation_2d_cell_numpy[t, y, x] = -1
            US_segmentation_2d_feature_numpy[t, y, x] = -1

        if dim == "3d":
            cell_indices_3d = where_replacement(US_segmentation_3d_cell_numpy == cell)
            features = -np.ones_like(US_segmentation_3d_feature_numpy)

            for t, z, y, x in cell_indices_3d:
                US_segmentation_3d_cell_numpy[t, z, y, x] = -1
                US_segmentation_3d_feature_numpy[t, z, y, x] = -1

                features[t, z, y, x] = US_segmentation_3d_feature_numpy[t, z, y, x]

        features = np.unique(features)[1:]
        US_features.drop(features, inplace=True)
        US_linking.drop(features, inplace=True)

    # convert back to data arrays
    US_segmentation_2d["Cell_Segmentation"] = (
        [*dims2d],
        US_segmentation_2d_cell_numpy,
    )
    US_segmentation_2d["Feature_Segmentation"] = (
        [*dims2d],
        US_segmentation_2d_feature_numpy,
    )
    if dim == "3d":
        US_segmentation_3d["Cell_Segmentation"] = (
            [*dims3d],
            US_segmentation_3d_cell_numpy,
        )
        US_segmentation_3d["Feature_Segmentation"] = (
            [*dims3d],
            US_segmentation_3d_feature_numpy,
        )

    # Recreate the analysis object
    analysis_object_filtered = Analysis_Object(
        analysis_object["tracking_xarray"],
        analysis_object["segmentation_xarray"],
        US_features,
        US_linking,
        US_segmentation_2d,
        US_segmentation_3d,
    )

    return analysis_object_filtered


def where_replacement(bool_arr):
    """
    A replacement function for numpy.where in attempts to speed up the process
    """
    import numpy as np
    from scipy.sparse import csr_matrix

    def compute_M(data):
        cols = np.arange(data.size)
        return csr_matrix(
            (cols, (data.ravel(), cols)), shape=(data.max() + 1, data.size)
        )

    def get_indices_sparse(data):
        M = compute_M(data)
        return [np.unravel_index(row.data, data.shape) for row in M]

    found_inds = get_indices_sparse(bool_arr)
    if len(found_inds) == 2:
        indices_to_replace = np.asarray(found_inds[1]).T
        return indices_to_replace
    elif len(found_inds) == 1:
        return []
    else:
        raise Exception("Maybe you dont actually know how this works")



import datetime
import re

import geopandas as gpd
import xarray as xr
from dateutil.parser import isoparse

from .TAMS import run


def goes_run_tams(
    goes_xarray: xr.DataArray,
    CONFIG: dict,
) -> tuple[gpd.GeoDataFrame, tuple]:
    """


    Parameters
    ----------
    goes_xarrayay : xarray.core.dataset.Dataset
        Xarray Dataset containing GOES data calculated from goes_load_netcdf(_iris).
    CONFIG : dict
        Default CoCoMET User CONFIG File.

    Returns
    -------
    ce : geopandas.geodataframe.GeoDataFrame
        A geopandas dataframe with the identified cloud elements.
    latlon_coord_system : tuple
        A tuple of the latitude and longitude coordinate arrays.

    """

    # make a copy of RAMS dataset to configure it to an acceptable format for TAMS
    goes_for_tams_copy = xr.Dataset({})

    tb = xr.DataArray(
        goes_xarray.values,
        dims=["Time", "south_north", "west_east"],
    )

    goes_for_tams_copy["ctt"] = tb.assign_attrs(
        {"long_name": "Brightness temperature", "units": "K"}
    )

    start_date = re.sub("^\\D*(\\d)", "\\1", goes_xarray["t"].attrs["units"])

    start_datetime = isoparse(start_date)
    start_year = start_datetime.year
    start_month = start_datetime.month
    start_day = start_datetime.day

    start_hour = start_datetime.hour
    start_minute = start_datetime.minute
    start_second = start_datetime.second

    time = []
    for minutes_since in goes_xarray["t"].values:

        new_temp_minute = int(start_minute + minutes_since)
        new_hour = int(start_hour + new_temp_minute // 60)
        new_second = int(start_second + new_temp_minute % 1)
        new_minute = new_temp_minute % 60 // 1

        new_time = datetime.datetime(
            start_year, start_month, start_day, new_hour, new_minute, new_second
        )
        time.append(new_time)

    # make the coordinates and rename the dimensions
    goes_for_tams_copy = goes_for_tams_copy.rename(
        {"Time": "time", "south_north": "lat", "west_east": "lon"}
    )  # rename the dimensions

    goes_for_tams_copy = goes_for_tams_copy.assign_coords(time=(["time"], time))
    goes_for_tams_copy = goes_for_tams_copy.assign_coords(
        lat=(["lat", "lon"], goes_xarray["lat"].data)
    )

    goes_for_tams_copy = goes_for_tams_copy.assign_coords(
        lon=(["lat", "lon"], goes_xarray["lon"].data)
    )

    # Had to manually return the lat/lon coordinate system to perform projections for analysis object
    ce, latlon_coord_system_2d = run(goes_for_tams_copy, **CONFIG["goes"]["tams"])

    # the latitude longitude coordinate systems are only in dimesions lat, lon. Add time dimension to them
    total_time = len(goes_for_tams_copy["time"])
    lat_coord_sys = latlon_coord_system_2d[0].expand_dims(
        dim={"time": range(total_time)}
    )
    lon_coord_sys = latlon_coord_system_2d[1].expand_dims(
        dim={"time": range(total_time)}
    )
    latlon_coord_system = (lat_coord_sys, lon_coord_sys)

    return ce, latlon_coord_system



import datetime

import geopandas as gpd
import xarray as xr

from .TAMS import run
from .wrf_calculate_products import (
    wrf_calculate_brightness_temp,
    wrf_calculate_precip_rate,
)


def wrf_run_tams(
    wrf_xarray: xr.Dataset,
    CONFIG: dict,
) -> tuple[gpd.GeoDataFrame, tuple]:
    """


    Parameters
    ----------
    wrf_xarray : xarray.core.dataset.Dataset
        Xarray Dataset containing WRF data calculated from wrf_load_netcdf(_iris).
    CONFIG : dict
        Default CoCoMET User CONFIG File.

    Returns
    -------
    ce : geopandas.geodataframe.GeoDataFrame
        A geopandas dataframe with the identified cloud elements.
    latlon_coord_system : tuple
        A tuple of the latitude and longitude coordinate arrays.

    """

    # make a copy of RAMS dataset to configure it to an acceptable format for TAMS
    wrf_for_tams_copy = xr.Dataset({})

    # if brightness temperature is already in wrf_xarray use it
    if "TB" not in wrf_xarray:
        tb = xr.DataArray(
            wrf_calculate_brightness_temp(wrf_xarray),
            dims=["Time", "south_north", "west_east"],
        )
    else:
        tb = wrf_xarray["TB"]

    wrf_for_tams_copy["ctt"] = tb.assign_attrs(
        {"long_name": "Brightness temperature", "units": "K"}
    )
    wrf_for_tams_copy["ctt"].chunk(wrf_xarray["XLAT"].chunksizes)
    wrf_xarray = wrf_xarray.assign(TB=wrf_for_tams_copy["ctt"])

    # # if precipitation rate is already in wrf_xarray use it
    # if "PR" not in wrf_xarray:
    #     pr = xr.DataArray(
    #         wrf_calculate_precip_rate(wrf_xarray),
    #         dims=["Time", "south_north", "west_east"],
    #     )

    # else:
    #     wrf_for_tams_copy["pr"] = wrf_xarray["PR"]

    # wrf_for_tams_copy["pr"] = pr.assign_attrs(
    #     {"long_name": "Precipitation rate", "units": "mm h-1"}
    # )
    # wrf_for_tams_copy["pr"].chunk(wrf_xarray["XLAT"].chunksizes)

    # format the times into a list of datetime objects

    time = []
    for t in wrf_xarray.XTIME.values:
        time_str = t.astype("datetime64[s]").astype(str)
        datetime_time = datetime.datetime.strptime(time_str, "%Y-%m-%dT%H:%M:%S")

        time.append(datetime_time)

    # make the coordinates and rename the dimensions
    wrf_for_tams_copy = wrf_for_tams_copy.rename(
        {"Time": "time", "south_north": "lat", "west_east": "lon"}
    )  # rename the dimensions

    wrf_for_tams_copy = wrf_for_tams_copy.assign_coords(time=(["time"], time))
    wrf_for_tams_copy = wrf_for_tams_copy.assign_coords(
        lat=(["time", "lat", "lon"], wrf_xarray["XLAT"].data)
    )
    wrf_for_tams_copy = wrf_for_tams_copy.assign_coords(
        lon=(["time", "lat", "lon"], wrf_xarray["XLONG"].data)
    )

    # Had to manually return the lat/lon coordinate system to perform projections for analysis object
    ce, latlon_coord_system = run(wrf_for_tams_copy, **CONFIG["wrf"]["tams"])

    return ce, latlon_coord_system



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 13 16:05:16 2024

@author: thahn
"""

# =============================================================================
# This defines the methods for running tobac on GOES data processed using goes_load.py
# =============================================================================

import logging

import geopandas as gpd
import iris
import iris.cube
import numpy as np
import tobac
import xarray as xr


def goes_tobac_feature_id(
    cube: iris.cube.Cube, CONFIG: dict
) -> gpd.GeoDataFrame | None:
    """


    Parameters
    ----------
    cube : iris.cube.Cube
        Iris cube containing the variable to be tracked.
    CONFIG : dict
        User configuration file.

    Returns
    -------
    goes_geopd : geopandas.geodataframe.GeoDataFrame
        Geodataframe containing all default tobac feature id outputs.

    """

    # Get horozontal spacings in km then convert to m
    res = float(cube.attributes["spatial_resolution"].split("km")[0]) * 1000

    dxy = tobac.get_spacings(cube, grid_spacing=res)[0]

    # Perform tobac feature identification and then convert to a geodataframe before returning
    goes_radar_features = tobac.feature_detection.feature_detection_multithreshold(
        cube, dxy=dxy, **CONFIG["goes"]["tobac"]["feature_id"]
    )

    if goes_radar_features is None:
        return None

    goes_geopd = gpd.GeoDataFrame(
        goes_radar_features,
        geometry=gpd.points_from_xy(
            goes_radar_features.longitude, goes_radar_features.latitude
        ),
        crs="EPSG:4326",
    )

    return goes_geopd


def goes_tobac_linking(
    cube: iris.cube.Cube, radar_features: gpd.GeoDataFrame, CONFIG: dict
) -> gpd.GeoDataFrame | None:
    """


    Parameters
    ----------
    cube : iris.cube.Cube
        Iris cube containing the variable to be tracked.
    radar_features : geopandas.geodataframe.GeoDataFrame
        tobac radar features from goes_tobac_feature_id output.
    CONFIG : dict
        User configuration file.

    Returns
    -------
    goes_geopd_tracks : geopandas.geodataframe.GeoDataFrame
        Geodataframe containing all default tobac feature id outputs.

    """

    if radar_features is None:
        return None

    # Mute tobac logging output
    logging.getLogger("trackpy").setLevel(level=logging.ERROR)

    # Get horozontal spacings in km then convert to m
    res = float(cube.attributes["spatial_resolution"].split("km")[0]) * 1000

    dxy = tobac.get_spacings(cube, grid_spacing=res)[0]

    # Get time spacing
    diffs = []
    for ii in range(cube.coord("time").points.shape[0] - 1):
        diffs.append(cube.coord("time").points[ii + 1] - cube.coord("time").points[ii])
    dt = np.nanmean(diffs) * 60

    # Do tracking then convert output dataframe to a geodataframe
    goes_tracks = tobac.linking_trackpy(
        radar_features, cube, dt=dt, dxy=dxy, **CONFIG["goes"]["tobac"]["linking"]
    )

    if goes_tracks is None:
        return None

    goes_geopd_tracks = gpd.GeoDataFrame(
        goes_tracks,
        geometry=gpd.points_from_xy(goes_tracks.longitude, goes_tracks.latitude),
        crs="EPSG:4326",
    )

    return goes_geopd_tracks


def goes_tobac_segmentation(
    cube: iris.cube.Cube,
    radar_features: gpd.GeoDataFrame,
    segmentation_type: str,
    CONFIG: dict,
    segmentation_height: None,
) -> tuple[xr.DataArray, gpd.GeoDataFrame]:
    """


    Parameters
    ----------
    cube : iris.cube.Cube
        Iris cube containing the variable to be tracked.
    radar_features : geopandas.geodataframe.GeoDataFrame
        tobac radar features from goes_tobac_feature_id output.
    CONFIG : dict
        User configuration file.

    Raises
    ------
    Exception
        Exception if using invalid tracking variable.

    Returns
    -------
    segment_xarray : xarray.core.dataarray.DataArray
        Xarray DataArray containing default tobac segmented data.
    segment_pd : geopandas.geodataframe.GeoDataFrame
        Geodataframe with ncells row.

    """

    if segmentation_type.lower() == "3d":
        return (None, None)

    if segmentation_height is not None:
        logging.warn("!=====segmentation_height Unused for GOES=====!")

    if radar_features is None:
        return (None, None)

    # Check tracking var
    if cube.name().lower() != "toa_brightness_temperature":
        raise Exception(
            f"!=====Invalid Tracking Variable. Your Cube Has: {cube.name().lower()}=====!"
        )

    # Get horozontal spacings in km then convert to m
    res = float(cube.attributes["spatial_resolution"].split("km")[0]) * 1000

    dxy = tobac.get_spacings(cube, grid_spacing=res)[0]

    # Perform the 2d segmentation and return the segmented cube and new geodataframe
    segment_cube, segment_features = tobac.segmentation_2D(
        radar_features, cube, dxy=dxy, **CONFIG["goes"]["tobac"]["segmentation_2d"]
    )

    # Convert iris cube to xarray and return
    return (xr.DataArray.from_iris(segment_cube), segment_features)



# =============================================================================
# This file contains the functions used to calculate additional values from rams output
# =============================================================================

import numpy as np
import xarray as xr
from tqdm import tqdm


# Calculate nearest item in list to given pivot
def find_nearest(array: np.ndarray, pivot) -> int:
    array = np.asarray(array)
    idx = (np.abs(array - pivot)).argmin()
    return idx


# TODO: This file needs a lot more explanation
def rams_calculate_brightness_temp(rams_xarray: xr.Dataset) -> xr.DataArray:
    """


    Parameters
    ----------
    rams_xarray : xarray.core.dataset.Dataset
        Xarray Dataset containing default RAMS values.

    Returns
    -------
    TB_xarray : xarray.core.dataarray.DataArray
        Xarray DataArray containing brightness temperature at each point and time--same dimension as input [K].

    """

    rams_xarray["TOA_OLR"] = (
        ["Time", "south_north", "west_east"],
        rams_xarray["LWUP"][:, -1, :, :].values,
    )  # the top of atmosphere surface radiation, needed to TB calculation
    TOA_OLR = rams_xarray["TOA_OLR"].values

    TB = np.zeros(TOA_OLR.shape)

    a = 1.228
    b = -1.106e-3
    sigma = 5.67e-8  # W m^-2 K^-4

    print("=====Calculating RAMS Brightness Temperatures=====")
    tf = (TOA_OLR[:, :, :] / sigma) ** 0.25
    TB[:, :, :] = (np.sqrt(4 * b * tf + a**2) - a) / (2 * b)

    TB_xarray = xr.DataArray(TB, dims=["Time", "south_north", "west_east"])

    TB_xarray = TB_xarray.chunk(rams_xarray["TOA_OLR"].chunksizes)

    return TB_xarray


def rams_calculate_precip_rate(
    rams_xarray: xr.Dataset, calculation_type: str | None = None
) -> xr.DataArray:
    """


    Parameters
    ----------
    rams_xarray : xarray.core.dataset.Dataset
        Xarray Dataset containing default RAMS values.
    calculation_type : str
        A string dictating which variables to use for the precipitation rate calculation

    Raises
    ------
        Exception if calculating precipitation rate with accumulation variables without ACCPR
        Exception if calculating precipitation rate with instantaneous rate variables without PCPRR
        Exception if calculating precipitation rate with volumetric instantaneous rate variables without PCPVR
        Exception if the calculation type is unknown

    Returns
    -------
    total2D_precip_rate_xarray : xarray.core.dataarray.DataArray
        Xarray DataArray of precipitation rate in mm/hr.

    """

    # If calculation type is none, look at the available variables in rams xarray and determine which scheme to use
    if calculation_type is None:
        if "ACCPR" in rams_xarray:
            calculation_type = "surface time averaged precipitation rate"

        elif "PCPRR" in rams_xarray:
            calculation_type = "surface instantaneous precipitation rate"

        elif "PCPVR" in rams_xarray:
            calculation_type = "volumetric instantaneous precipitation rate"

        else:
            raise Exception("No precipitation variables found in data")

    # based on the calculation type, calculate precipitation rate
    if calculation_type == "surface time averaged precipitation rate":
        accumulated_rainfall_vars = [
            "ACCPR",
            "ACCPP",
            "ACCPS",
            "ACCPA",
            "ACCPG",
            "ACCPH",
            "ACCPD",
            "ACONPR",
            "ACCPIP",
            "ACCPIC",
            "ACCPID",
        ]

        if "ACCPR" not in rams_xarray:
            raise Exception(
                "If calculating precipitation with accumulation variables, ACCPR must be present"
            )

        total_accumulation = rams_xarray["ACCPR"]
        for avar in accumulated_rainfall_vars[1:]:
            if avar in rams_xarray:
                total_accumulation += rams_xarray[avar]

        pr = np.zeros_like(total_accumulation)

        for tt in tqdm(
            range(pr.shape[0] - 1),
            desc="=====Calculating RAMS Precipitation Rate=====",
            total=pr.shape[0] - 1,
        ):
            pr[tt] = (
                total_accumulation[tt] - total_accumulation[tt - 1]
            ) / rams_xarray.DT  # in kg / m^2 / s

        # pr = pr / 1000 * 1000 # convert to mm / s, (kg / m^2 / s) * (1/1000 m^3/kg) * (1000 mm / m)
        pr = xr.DataArray(pr, dims=["Time", "south_north", "west_east"]).chunk(
            rams_xarray["TOPT"].chunksizes
        )

    elif calculation_type == "surface instantaneous precipitation rate":
        instantaneous_rainfall_vars = [
            "PCPRR",
            "PCPRP",
            "PCPRS",
            "PCPRA",
            "PCPRG",
            "PCPRH",
            "PCPRD",
            "CONPRR",
            "PCPRIP",
            "PCPRIC",
            "PCPRID",
        ]

        if "PCPRR" not in rams_xarray:
            raise Exception(
                "If calculating precipitation with instantaneous variables, PCPRR must be present"
            )

        pr = rams_xarray["PCPRR"]
        for avar in instantaneous_rainfall_vars[1:]:
            if avar in rams_xarray:
                pr += rams_xarray[avar]

        pr = pr.chunk(rams_xarray["TOPT"].chunksizes)

    elif calculation_type == "volumetric instantaneous precipitation rate":
        instantaneous_rainfall_vars = [
            "PCPVR",
            "PCPVP",
            "PCPVS",
            "PCPVA",
            "PCPVG",
            "PCPVH",
            "PCPVD",
            "CONPRR",
            "PCPVIP",
            "PCPVIC",
            "PCPVID",
        ]

        if "PCPVR" not in rams_xarray:
            raise Exception(
                "If calculating precipitation with the volumetric instantaneous variables, PCPVR must be present"
            )

        pr = rams_xarray["PCPVR"]
        for avar in instantaneous_rainfall_vars[1:]:
            if avar in rams_xarray:
                pr += rams_xarray[avar]

        pr = pr.chunk(rams_xarray["PCPVR"].chunksizes)

    else:
        raise Exception("Unknown calculation type for precipitation rate calculation")
    return pr


# Why does this function exist?
def rams_calculate_wa(rams_xarray: xr.Dataset) -> xr.DataArray:
    """


    Parameters
    ----------
    rams_xarray : xarray.core.dataset.Dataset
        xarray Dataset containing default RAMS values.

    Returns
    -------
    wa : xarray.core.dataarray.DataArray
        Xarray DataArray of vertical wind components at mass points.

    """

    wa = rams_xarray["WC"]  # rams winds are unstaggered in the header file already
    wa = wa.assign_attrs(
        {
            "units": "m s-1",
            "coordinates": "GLON GLAT Times",
            "description": "updraft velocity",
            "MemoryOrder": "XYZ",
        }
    )
    return wa


# Citation needed for this methodology
def rams_calculate_reflectivity(rams_xarray: xr.Dataset) -> xr.DataArray:
    """


    Parameters
    ----------
    rams_xarray : xarray.core.dataset.Dataset
        Xarray Dataset containing default RAMS values.

    Returns
    -------
    dBZ: xarray.core.dataarray.DataArray
        Xarray DataArray of calculated reflectivity values.

    """

    # constants defined by RAMS
    p00 = 1.0e5  # Pa
    rgas = 287  # J/gk/K
    cp = 1004  # J/kg/K
    cpor = cp / rgas

    # find temperature, pressure, and grid point density
    T_pot = rams_xarray["THETA"]  # potential temperature
    PI = rams_xarray["PI"]
    P = ((PI / cp) ** cpor) * p00 * 0.01  # convert pressure to mb
    T = T_pot * PI / cp  # temperature in K
    dens = (P * 100) / (T * rgas)

    # mass coefficients defined by RAMS
    alpha_mr = 524.0  # rain mass coeff
    alpha_mg = 157.0  # graupel mass coeff
    alpha_mh = 471.0  # hail mass coeff
    alpha_mp = 110.8  # pris mass coeff
    alpha_ms = 2.739e-3  # snow mass coeff
    alpha_ma = 0.496  # aggregates mass coeff

    # need mixing ratios and concentrations and ensure they are above 0
    rainMix = xr.where(rams_xarray["RRP"] < 0, 0, rams_xarray["RRP"])
    pris_iceMix = xr.where(rams_xarray["RPP"] < 0, 0, rams_xarray["RPP"])
    snowMix = xr.where(rams_xarray["RSP"] < 0, 0, rams_xarray["RSP"])
    aggregatesMix = xr.where(rams_xarray["RAP"] < 0, 0, rams_xarray["RAP"])
    graupelMix = xr.where(rams_xarray["RGP"] < 0, 0, rams_xarray["RGP"])
    hailMix = xr.where(rams_xarray["RHP"] < 0, 0, rams_xarray["RHP"])
    rainConc = xr.where(rams_xarray["CRP"] < 0, 0, rams_xarray["CRP"])
    pris_iceConc = xr.where(rams_xarray["CPP"] < 0, 0, rams_xarray["CPP"])
    snowConc = xr.where(rams_xarray["CSP"] < 0, 0, rams_xarray["CSP"])
    aggregatesConc = xr.where(rams_xarray["CAP"] < 0, 0, rams_xarray["CAP"])
    graupelConc = xr.where(rams_xarray["CGP"] < 0, 0, rams_xarray["CGP"])
    hailConc = xr.where(rams_xarray["CHP"] < 0, 0, rams_xarray["CHP"])

    # define the gamma shape parameters
    gamma = rams_xarray.gnu  # in order: cld rain pris snow aggr graup hail driz

    dicToIterateThrough = {
        "rain": [alpha_mr, rainMix, rainConc, gamma[1]],
        "graupel": [alpha_mg, graupelMix, graupelConc, gamma[5]],
        "pris_ice": [alpha_mp, pris_iceMix, pris_iceConc, gamma[2], 2.91],
        "snow": [alpha_ms, snowMix, snowConc, gamma[3], 1.74],
        "hail": [alpha_mh, hailMix, hailConc, gamma[6]],
        "aggregates": [alpha_ma, aggregatesMix, aggregatesConc, gamma[4], 2.4],
    }

    # define an array of the reflectivity and iteratively add to it
    dbz_total = np.zeros_like(rainMix)
    Z_total = np.zeros_like(rainMix)

    # thresholds for the mixing ratios and concentration numbers
    q = 1e-10  # kg / kg
    qn = 1e-3  # (# / kg)

    for key in tqdm(
        dicToIterateThrough.keys(),
        desc="=====Calculating RAMS Reflectivity=====",
        total=len(dicToIterateThrough.keys()),
    ):
        alpha = dicToIterateThrough[key][0]
        mixRatio = dicToIterateThrough[key][1]
        conc = dicToIterateThrough[key][2]
        gsp = dicToIterateThrough[key][3]

        F_gnu1 = (5.0 + gsp) * (4.0 + gsp) * (3.0 + gsp)
        F_gnu2 = (2.0 + gsp) * (1.0 + gsp) * gsp
        F_gnu = F_gnu1 / F_gnu2

        # if the mixing ratio or number concentration is below a threshold there is no contribution to reflectivity
        mixRatiobool = mixRatio > q
        concbool = conc > qn
        FullBool = mixRatiobool * concbool

        if key in ["pris_ice", "snow", "aggregates"]:
            factor = dicToIterateThrough[key][4]
            M = mixRatio / conc
            D = (M / alpha) ** (1 / factor)
            alpha = M / (D**3)
        tmp0 = mixRatio / alpha
        tmp1 = (tmp0**2) * dens * F_gnu
        tmp2 = (tmp1 / conc) * 1.0e18
        tmp2 = np.where(FullBool, tmp2, 0)

        Z_total += tmp2

    Z_total = Z_total.clip(0.001, 1e99)

    dbz_total = 10 * np.log10(Z_total)

    dbz_total_xarray = xr.DataArray(
        dbz_total, dims=["Time", "bottom_top", "south_north", "west_east"]
    )

    # Assign attributes
    dBZ = dbz_total_xarray.assign_attrs(
        {
            "FieldType": 104,
            "MemoryOrder": "XYZ",
            "description": "radar reflectivity",
            "units": "dBZ",
            "stagger": "",
            "coordinates": "Times GLON GLAT",
        }
    )

    return dBZ.chunk(T.chunksizes)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 24 12:17:45 2024

@author: thahn
"""

# =============================================================================
# This defines the primary functionality of the analysis module for CoCoMET. Including the primary get_var function which works similar to wrf-python's getvar
# =============================================================================

from CoCoMET.analysis.analysis_object import Analysis_Object

from .calculate_arm_products import calculate_convective_indices, extract_arm_product
from .calculate_bulk_cell_statistics import (
    calculate_area,
    calculate_cell_growth,
    calculate_irregularity,
    calculate_max_intensity,
    calculate_perimeter,
    calculate_surface_area,
    calculate_var_max_height,
    calculate_velocity,
    calculate_volume,
)
from .merge_split_detection import merge_split


def calc_var(analysis_object: Analysis_Object, var: str, **args: dict):
    """


    Parameters
    ----------
    analysis_object : Analysis_Object
        The output of CoCoMET_start and/or an object generated by create_analysis_object following CoCoMET-US Specification 1.x.
    var : str
        Which variable to calculate. See list below.
    **args : dict
        The additional arguments of indeterminate length required for calculating request variables.

    Raises
    ------
    ValueError
        Exception if invalid variable name.

    Returns
    -------
    Type depends on variable
        N/A.

    Valid Variables
    ---------------
        "var_max_height",
        "max_intensity",
        "area",
        "volume",
        "velocity",
        "perimeter",
        "surface_area",
        "cell_growth",
        "irregularity",
        "merge_split",
        "arm",
        "convective_indices"

    """

    # Map the correct functions to the proper variables. This is a list of all the calculatable variables as well.
    variable_call_mechanism = {
        "var_max_height": calculate_var_max_height,
        "max_intensity": calculate_max_intensity,
        "area": calculate_area,
        "volume": calculate_volume,
        "velocity": calculate_velocity,
        "perimeter": calculate_perimeter,
        "surface_area": calculate_surface_area,
        "cell_growth": calculate_cell_growth,
        "irregularity": calculate_irregularity,
        "merge_split": merge_split,
        "arm": extract_arm_product,
        "convective_indices": calculate_convective_indices,
    }

    # Check for valid variables
    if var.lower() not in variable_call_mechanism:
        raise ValueError(
            f"!=====Invalid Variable Requested. You Entered: {var.lower()}=====!"
        )

    # Call the proper function and return its output
    return variable_call_mechanism[var.lower()](
        analysis_object=analysis_object.return_analysis_dictionary(),
        **args,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# pylint: skip-file
"""
Created on Tue Jul  9 13:02:43 2024

@author: thahn

ADAPTED FROM wrfcube package: 

BSD 3-Clause License

Copyright (c) 2019, Max Heikenfeld
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import vincenty
from iris import Constraint, coord_systems, coords
from iris.coords import AuxCoord
from iris.util import promote_aux_coord_to_dim_coord
from numpy import arange, array, transpose


def load(mesonh_xarray, variable, filename):
    array = mesonh_xarray[variable]
    variable_dimensions = array.dims
    attributes = mesonh_xarray.attrs
    cube = array.to_iris()
    coord_system = make_coord_system(attributes)

    dy, dx = guess_horizontal_spacing(mesonh_xarray, filename)

    for dim in range(len(variable_dimensions)):
        if variable_dimensions[dim] == "x":
            west_east = make_westeast_coord(dx, mesonh_xarray.x.size)
            cube.add_dim_coord(west_east, dim)
            projection_x_coord = make_x_coord(
                dx, mesonh_xarray.x.size, coord_system=coord_system
            )
            cube.add_aux_coord(projection_x_coord, dim)
            x_coord = AuxCoord(cube.coord("west_east").points, long_name="x", units=1)
            cube.add_aux_coord(x_coord, data_dims=cube.coord_dims("west_east"))
        elif variable_dimensions[dim] == "y":
            south_north = make_southnorth_coord(dy, mesonh_xarray.y.size)
            cube.add_dim_coord(south_north, dim)
            projection_y_coord = make_y_coord(
                dy, mesonh_xarray.y.size, coord_system=coord_system
            )
            cube.add_aux_coord(projection_y_coord, dim)
            y_coord = AuxCoord(cube.coord("south_north").points, long_name="y", units=1)
            cube.add_aux_coord(y_coord, data_dims=cube.coord_dims("south_north"))
        elif variable_dimensions[dim] == "z":
            bottom_top = make_bottom_top_coordinate(mesonh_xarray.z.size)
            cube.add_dim_coord(bottom_top, dim)
            model_level_number = make_model_level_number_coordinate(
                mesonh_xarray.z.size
            )
            cube.add_aux_coord(model_level_number, dim)

    if "time" in [coord.name() for coord in cube.coords()]:
        promote_aux_coord_to_dim_coord(cube, "time")
        cube.coord("time").attributes = {}

    # change latitude and longitude coordinates to  2D fields (fine andm ore consistent with other models for all static WRF Simulations)
    if "lat" in [coord.name() for coord in cube.coords()]:
        latitude_coord = cube.extract(
            Constraint(
                time=cube.coord("time").units.num2date(cube.coord("time").points[0])
            )
        ).coord("lat")
        latitude_coord.rename("latitude")
        xlat_dims = list(cube.coord_dims("lat"))
        time_dim = cube.coord_dims("time")[0]
        xlat_dims.remove(time_dim)
        data_dims = tuple(xlat_dims)
        cube.add_aux_coord(latitude_coord, data_dims=data_dims)
        cube.remove_coord("lat")
    if "lon" in [coord.name() for coord in cube.coords()]:
        longitude_coord = cube[0].coord("lon")
        longitude_coord.rename("longitude")
        xlong_dims = list(cube.coord_dims("lon"))
        time_dim = cube.coord_dims("time")[0]
        xlong_dims.remove(time_dim)
        data_dims = tuple(xlong_dims)
        cube.add_aux_coord(longitude_coord, data_dims=data_dims)
        cube.remove_coord("lon")

    return cube


def guess_horizontal_spacing(mesonh_xarray, filename):
    try:
        # Try to guess dimension from file name
        dis_value = filename.split("m")[0]

        # If is in kilometers, convert to meters
        if "k" in dis_value:
            dis_value = float(dis_value.replace("k", "")) * 1000

        else:
            dis_value = float(dis_value)

        return (dis_value, dis_value)

    except ValueError:
        print(
            "!=====Non-Default MesoNH Filename Found, Estimating Distance Instead=====!"
        )

        # Guess x dimension by finding distance between two points offset by one x value
        x_dis = (
            vincenty.vincenty(
                (mesonh_xarray.lat[0, 0, 0].values, mesonh_xarray.lon[0, 0, 0].values),
                (mesonh_xarray.lat[0, 0, 1].values, mesonh_xarray.lon[0, 0, 1].values),
            )
            * 1000
        )
        # Guess y dimension by finding distance between two points offset by one y value
        y_dis = (
            vincenty.vincenty(
                (mesonh_xarray.lat[0, 0, 0].values, mesonh_xarray.lon[0, 0, 0].values),
                (mesonh_xarray.lat[0, 1, 0].values, mesonh_xarray.lon[0, 1, 0].values),
            )
            * 1000
        )

        # Round to nearest tens place
        x_dis = 10 * round(x_dis / 10)
        y_dis = 10 * round(y_dis / 10)

        return (y_dis, x_dis)


def make_coord_system(attributes):
    return None

    #    :CEN_LAT = -3.212929f ;
    # 		:CEN_LON = -60.59799f ;
    # 		:TRUELAT1 = 0.f ;
    # 		:TRUELAT2 = -5.f ;
    # 		:MOAD_CEN_LAT = -3.212929f ;
    # 		:STAND_LON = -60.f ;
    # 		:POLE_LAT = 90.f ;
    # 		:POLE_LON = 0.f ;
    # 		:GMT = 0.f ;
    # 		:JULYR = 2014 ;
    # 		:JULDAY = 244 ;
    # 		:MAP_PROJ = 1 ;
    # 		:MAP_PROJ_CHAR = "Lambert Conformal" ;
    MAP_PROJ_CHAR = attributes["MAP_PROJ_CHAR"]
    MAP_PROJ = attributes["MAP_PROJ"]

    # cartesian coordinate system (idealized simulations):
    if MAP_PROJ_CHAR == "Cartesian" and MAP_PROJ == 0:
        coord_system = None

    # lambert Conformal system (idealized simulations):
    if MAP_PROJ_CHAR == "Lambert Conformal" and MAP_PROJ == 1:
        CEN_LON = attributes["CEN_LON"]
        TRUELAT1 = attributes["TRUELAT1"]
        TRUELAT2 = attributes["TRUELAT2"]
        MOAD_CEN_LAT = attributes["MOAD_CEN_LAT"]
        # STAND_LON = attributes["STAND_LON"]
        # POLE_LAT = attributes["POLE_LAT"]
        # POLE_LON = attributes["POLE_LON"]
        coord_system = coord_systems.LambertConformal(
            central_lat=MOAD_CEN_LAT,
            central_lon=CEN_LON,
            false_easting=0.0,
            false_northing=0.0,
            secant_latitudes=(TRUELAT1, TRUELAT2),
        )
    return coord_system


def make_westeast_coord(DX, WEST_EAST_PATCH_END_UNSTAG):
    WEST_EAST = arange(0, WEST_EAST_PATCH_END_UNSTAG)
    west_east = coords.DimCoord(
        WEST_EAST,
        standard_name=None,
        long_name="west_east",
        var_name="west_east",
        units="1",
        bounds=None,
        attributes=None,
        coord_system=None,
        circular=False,
    )
    return west_east


def make_westeast_stag_coord(DX, WEST_EAST_PATCH_END_STAG):
    WEST_EAST_U = arange(0, WEST_EAST_PATCH_END_STAG)
    west_east_stag = coords.DimCoord(
        WEST_EAST_U,
        standard_name=None,
        long_name="west_east_stag",
        var_name="west_east_stag",
        units="1",
        bounds=None,
        attributes=None,
        coord_system=None,
        circular=False,
    )
    return west_east_stag


def make_southnorth_coord(DY, SOUTH_NORTH_PATCH_END_UNSTAG):
    # SOUTH_NORTH_PATCH_END_UNSTAG=attributes["SOUTH-NORTH_PATCH_END_UNSTAG"]
    SOUTH_NORTH = arange(0, SOUTH_NORTH_PATCH_END_UNSTAG)
    south_north = coords.DimCoord(
        SOUTH_NORTH,
        standard_name=None,
        long_name="south_north",
        var_name="south_north",
        units="1",
        bounds=None,
        attributes=None,
        coord_system=None,
        circular=False,
    )
    return south_north


def make_southnorth_stag_coord(DY, SOUTH_NORTH_PATCH_END_STAG):
    SOUTH_NORTH_V = arange(0, SOUTH_NORTH_PATCH_END_STAG)
    south_north_stag = coords.DimCoord(
        SOUTH_NORTH_V,
        standard_name=None,
        long_name="south_north_stag",
        var_name="south_north_stag",
        units="1",
        bounds=None,
        attributes=None,
        coord_system=None,
        circular=False,
    )
    return south_north_stag


def make_bottom_top_coordinate(BOTTOM_TOP_PATCH_END_UNSTAG):
    BOTTOM_TOP = arange(0, BOTTOM_TOP_PATCH_END_UNSTAG)
    bottom_top = coords.DimCoord(
        BOTTOM_TOP,
        standard_name=None,
        long_name="altitude",
        var_name="altitude",
        units="m",
        bounds=None,
        attributes=None,
        coord_system=None,
        circular=False,
    )
    return bottom_top


def make_bottom_top_stag_coordinate(BOTTOM_TOP_PATCH_END_STAG):
    BOTTOM_TOP_W = arange(0, BOTTOM_TOP_PATCH_END_STAG)
    bottom_top_stag = coords.DimCoord(
        BOTTOM_TOP_W,
        standard_name=None,
        long_name="altitude_stag",
        var_name="altitude_stag",
        units="m",
        bounds=None,
        attributes=None,
        coord_system=None,
        circular=False,
    )
    return bottom_top_stag


def make_model_level_number_coordinate(BOTTOM_TOP_PATCH_END):
    MODEL_LEVEL_NUMBER = arange(0, BOTTOM_TOP_PATCH_END)
    model_level_number = coords.AuxCoord(
        MODEL_LEVEL_NUMBER, standard_name="model_level_number", units="1"
    )
    return model_level_number


def make_x_coord(DX, WEST_EAST_PATCH_END_UNSTAG, coord_system):
    X = DX * (arange(0, WEST_EAST_PATCH_END_UNSTAG) + 0.5)
    bounds = transpose(
        array(
            [
                DX * (arange(0, WEST_EAST_PATCH_END_UNSTAG)),
                DX * (arange(0, WEST_EAST_PATCH_END_UNSTAG) + 1),
            ]
        )
    )
    x_coord = coords.AuxCoord(
        X,
        standard_name="projection_x_coordinate",
        long_name="x",
        var_name="x",
        units="m",
        bounds=bounds,
        attributes=None,
        coord_system=coord_system,
    )
    # x_coord.add_dim_coord(west_east,0)
    return x_coord


def make_x_stag_coord(DX, WEST_EAST_PATCH_END_STAG, coord_system=None):
    X_U = DX * (arange(0, WEST_EAST_PATCH_END_STAG) - 1)
    x_stag_coord = coords.AuxCoord(
        X_U,
        standard_name="projection_x_coordinate",
        long_name="x",
        var_name="x",
        units="m",
        bounds=None,
        attributes=None,
        coord_system=coord_system,
    )
    # x_stag_coord.add_dim_coord(west_east_stag,0)
    return x_stag_coord


def make_y_coord(DY, SOUTH_NORTH_PATCH_END_UNSTAG, coord_system=None):
    Y = DY * (arange(0, SOUTH_NORTH_PATCH_END_UNSTAG) + 0.5)
    bounds = transpose(
        array(
            [
                DY * (arange(0, SOUTH_NORTH_PATCH_END_UNSTAG)),
                DY * (arange(0, SOUTH_NORTH_PATCH_END_UNSTAG) + 1),
            ]
        )
    )
    y_coord = coords.AuxCoord(
        Y,
        standard_name="projection_y_coordinate",
        long_name="y",
        var_name="y",
        units="m",
        bounds=bounds,
        attributes=None,
        coord_system=coord_system,
    )
    # y_coord.add_dim_coord(south_north,0)
    return y_coord


def make_y_stag_coord(DY, SOUTH_NORTH_PATCH_END_STAG, coord_system=None):
    Y_V = DY * (arange(0, SOUTH_NORTH_PATCH_END_STAG) - 1)
    y_stag_coord = coords.AuxCoord(
        Y_V,
        standard_name="projection_y_coordinate",
        long_name="y",
        var_name="y",
        units="m",
        bounds=None,
        attributes=None,
        coord_system=coord_system,
    )
    # y_stag_coord.add_dim_coord(south_north_stag,0)
    return y_stag_coord



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 21 18:01:16 2024

@author: thahn
"""

# =============================================================================
# Takes in a filepath containing WRF netCDF data and converts it to a netcdf dataset and/or an iris cube for use in trackers
# =============================================================================

import datetime
import glob
import logging
import os

import iris
import iris.cube
import numpy as np
import vincenty
import xarray as xr

from .mesonh_calculate_products import (
    mesonh_calculate_agl_z,
    mesonh_calculate_brightness_temp,
    mesonh_calculate_reflectivity,
)
from .mesonhcube import load


def guess_horizontal_spacing(
    mesonh_xarray: xr.Dataset, filename: str
) -> tuple[float, float]:
    """
    This functions attempts to find the horizontal spacing of MesoNH data

    Parameters
    ----------
    mesonh_xarray : xarray.core.dataset.Dataset
        Standard mesonh xarray dataset.
    filename : str
        The name of a mesonh input file.

    Returns
    -------
    dy : float
        Estimated y spacing (m).
    dx : float
        Estimated x spacing (m).
    """

    try:
        # Try to guess dimension from file name
        dis_value = filename.split("m")[0]

        # If is in kilometers, convert to meters
        if "k" in dis_value:
            dis_value = float(dis_value.replace("k", "")) * 1000

        else:
            dis_value = float(dis_value)

        return (dis_value, dis_value)

    except ValueError:
        print(
            "!=====Non-Default MesoNH Filename Found, Estimating Distance Instead=====!"
        )

        # Guess x dimension by finding distance between two points offset by one x value
        x_dis = (
            vincenty.vincenty(
                (mesonh_xarray.lat[0, 0, 0].values, mesonh_xarray.lon[0, 0, 0].values),
                (mesonh_xarray.lat[0, 0, 1].values, mesonh_xarray.lon[0, 0, 1].values),
            )
            * 1000
        )
        # Guess y dimension by finding distance between two points offset by one y value
        y_dis = (
            vincenty.vincenty(
                (mesonh_xarray.lat[0, 0, 0].values, mesonh_xarray.lon[0, 0, 0].values),
                (mesonh_xarray.lat[0, 1, 0].values, mesonh_xarray.lon[0, 1, 0].values),
            )
            * 1000
        )

        # Round to nearest tens place
        x_dis = 10 * round(x_dis / 10)
        y_dis = 10 * round(y_dis / 10)

        return (y_dis, x_dis)


def mesonh_load_netcdf_iris(
    filepath: str, tracking_var: str, CONFIG: dict
) -> tuple[iris.cube.Cube, xr.Dataset]:
    """


    Parameters
    ----------
    filepath : str
        Glob style path to MesoNH files (i.e. ./data/MesoNH/500m*).
    tracking_var : str
        ["dbz","tb","wa", "pr", ...], variable which is going to be used for tracking
    CONFIG : dict
        User configuration file.

    Raises
    ------
    Exception
        Exception if entered invalid tracking variable.
        ValueError if the simulation times are not continuous.

    Returns
    -------
    cube : iris.cube.Cube
        Iris cube containing tracking variable.
    mesonh_xarray : xarray.core.dataset.Dataset
        Xarray dataset containing merged MesoNH data.

    """

    # Get one filename for guessing spacing
    filename = [os.path.basename(x) for x in glob.glob(filepath)][0]
    mesonh_xarray = xr.open_mfdataset(
        filepath, coords="all", concat_dim="time", combine="nested"
    )

    if "mesonh" in CONFIG:
        # Subset time based on user inputs
        if (
            "min_frame_index" in CONFIG["mesonh"]
            or "max_frame_index" in CONFIG["mesonh"]
        ):
            min_frame = (
                CONFIG["mesonh"]["min_frame_index"]
                if "min_frame_index" in CONFIG["mesonh"]
                else 0
            )
            max_frame = (
                CONFIG["mesonh"]["max_frame_index"] + 1
                if "max_frame_index" in CONFIG["mesonh"]
                else mesonh_xarray.dims["time"]
            )

            mesonh_xarray = mesonh_xarray.isel(
                time=np.arange(min_frame, max_frame), drop=True
            )

    else:
        raise Exception("""!=====CONFIG Missing "mesonh" field=====!""")

    # Get the time differences
    datetime_times = [
        datetime.datetime.strptime(str(time), "%Y-%m-%dT%H:%M:%S.%f000")
        for time in mesonh_xarray["time"].values
    ]
    dt_datetimes = np.diff(datetime_times)
    dt_seconds = [s.total_seconds() for s in dt_datetimes]

    if len(np.unique(dt_seconds)) > 1:
        raise ValueError("Simulation times are inconsistent")
    else:
        mesonh_xarray.attrs["DT"] = np.unique(dt_seconds)[0]

    # Correct for 360 degree lat/lon system by subtracting 360 from values exceeding 180 degrees...correction for latitude may not be necessary
    mesonh_xarray = mesonh_xarray.assign_coords(
        lat=mesonh_xarray.lat.where(mesonh_xarray.lat <= 180, lambda lat: lat - 360),
        lon=mesonh_xarray.lon.where(mesonh_xarray.lon <= 180, lambda lon: lon - 360),
    )
    mesonh_xarray = mesonh_xarray.unify_chunks()

    # mesonh should be in CONFIG from previous check, check bounds after lat/lon correction
    if "bounds" in CONFIG["mesonh"]:
        # If it is idealized data, print a warning
        if "is_idealized" in CONFIG["mesonh"]:
            if CONFIG["mesonh"]["is_idealized"]:
                logging.warning("!=====Setting bounds for idealized data=====!")

        mask_lon = (mesonh_xarray.lon >= CONFIG["mesonh"]["bounds"][0]) & (
            mesonh_xarray.lon <= CONFIG["mesonh"]["bounds"][1]
        )
        mask_lat = (mesonh_xarray.lat >= CONFIG["mesonh"]["bounds"][2]) & (
            mesonh_xarray.lat <= CONFIG["mesonh"]["bounds"][3]
        )

        mesonh_xarray = mesonh_xarray.where(mask_lon & mask_lat, drop=True)

    dx, dy = guess_horizontal_spacing(mesonh_xarray, filename)

    # Add projection x and y coordinates to MesoNH
    proj_y_values = dy * (np.arange(0, mesonh_xarray.y.size) + 0.5)
    proj_x_values = dx * (np.arange(0, mesonh_xarray.x.size) + 0.5)

    mesonh_xarray["PROJY"] = ("y", proj_y_values)
    mesonh_xarray["PROJX"] = ("x", proj_x_values)

    if tracking_var.lower() == "dbz":
        mesonh_reflectivity = mesonh_calculate_reflectivity(mesonh_xarray)

        mesonh_xarray["DBZ"] = mesonh_reflectivity
        cube = load(mesonh_xarray, "DBZ", filename)

        # add correct altitude based off of average height at each height index
        ht = mesonh_calculate_agl_z(mesonh_xarray)

        correct_alts = [np.mean(h.values) for h in ht]
        cube.coord("altitude").points = correct_alts

        # Add altitude field for easier processing later
        mesonh_xarray["DBZ"] = mesonh_xarray["DBZ"].assign_coords(
            altitude=("z", correct_alts)
        )

        # Add altitudes as another data variable
        mesonh_xarray["altitudes"] = (["z"], correct_alts)

    elif tracking_var.lower() == "tb":
        # Brightness temperature is only 2d so no heights needed
        mesonh_xarray["TB"] = (
            ["time", "y", "x"],
            mesonh_calculate_brightness_temp(mesonh_xarray),
        )
        mesonh_xarray["TB"].attrs["units"] = "K"

        # Adjust dask chunks
        mesonh_xarray["TB"] = mesonh_xarray["TB"].chunk(
            mesonh_xarray["LWup_TOA"].chunksizes
        )
        cube = load(mesonh_xarray, "TB", filename)

    elif tracking_var.lower() == "wa":
        # Get updraft velocity at mass points (maybe?)
        mesonh_wa = mesonh_xarray.w

        mesonh_xarray["WA"] = mesonh_wa
        cube = load(mesonh_xarray, "WA", filename)

        # Add correct altitude based off of average height at each height index
        ht = mesonh_calculate_agl_z(mesonh_xarray)

        correct_alts = [np.mean(h.values) for h in ht]
        cube.coord("altitude").points = correct_alts

        # Add altitude field for easier processing later
        mesonh_xarray["WA"] = mesonh_xarray["WA"].assign_coords(
            altitude=("z", correct_alts)
        )

        # Add altitudes as another data variable
        mesonh_xarray["altitudes"] = (["z"], correct_alts)

    elif tracking_var.lower() == "pr":
        # precipitation rate is only 2d so no heights needed
        mesonh_xarray["PR"] = mesonh_xarray["pcp_rate"]

        mesonh_xarray["PR"].attrs["units"] = "mm/hour"

        # Adjust dask chunks
        mesonh_xarray["PR"] = mesonh_xarray["PR"].chunk(
            mesonh_xarray["LWup_TOA"].chunksizes
        )
        cube = load(mesonh_xarray, "PR", filename)

    else:
        # If not any of the above, try using user inputed value
        try:
            var_values = mesonh_xarray[tracking_var]
            cube = load(mesonh_xarray, tracking_var, filename)

            if len(var_values.shape) == 4:
                # add correct altitude based off of average height at each height index
                ht = mesonh_calculate_agl_z(mesonh_xarray)

                correct_alts = [np.mean(h.values) for h in ht]
                cube.coord("altitude").points = correct_alts

                # Add altitude field for easier processing later
                mesonh_xarray[tracking_var] = mesonh_xarray[tracking_var].assign_coords(
                    altitude=("z", correct_alts)
                )

                # Add altitudes as another data variable
                mesonh_xarray["altitudes"] = (["z"], correct_alts)

        except:
            raise Exception(
                f"!=====Invalid Tracking Variable. You Entered: {tracking_var}=====!"
            )

    return (cube, mesonh_xarray.unify_chunks())


def mesonh_load_netcdf(filepath: str, tracking_var: str, CONFIG: dict) -> xr.Dataset:
    """


    Parameters
    ----------
    filepath : str
        Glob style path to MesoNH files (i.e. ./data/MesoNH/500m*).
    tracking_var : str
        ["dbz","tb","wa", "pr", ...], variable which is going to be used for tracking
    CONFIG : dict
        User configuration file.

    Raises
    ------
    Exception
        Exception if entered invalid tracking variable.

    Returns
    -------
    mesonh_xarray : xarray.core.dataset.Dataset
        Xarray dataset containing merged MesoNH data.

    """

    # Get one filename for guessing spacing
    filename = [os.path.basename(x) for x in glob.glob(filepath)][0]

    mesonh_xarray = xr.open_mfdataset(
        filepath, coords="all", concat_dim="time", combine="nested"
    )

    if "mesonh" in CONFIG:
        # Subset time based on user inputs
        if (
            "min_frame_index" in CONFIG["mesonh"]
            or "max_frame_index" in CONFIG["mesonh"]
        ):
            min_frame = (
                CONFIG["mesonh"]["min_frame_index"]
                if "min_frame_index" in CONFIG["mesonh"]
                else 0
            )
            max_frame = (
                CONFIG["mesonh"]["max_frame_index"] + 1
                if "max_frame_index" in CONFIG["mesonh"]
                else mesonh_xarray.dims["time"]
            )

            mesonh_xarray = mesonh_xarray.isel(
                time=np.arange(min_frame, max_frame), drop=True
            )

    else:
        raise Exception("""!=====CONFIG Missing "mesonh" field=====!""")

    # Correct for 360 degree lat/lon system by subtracting 360 from values exceeding 180 degrees...correction for latitude may not be necessary
    mesonh_xarray = mesonh_xarray.assign_coords(
        lat=mesonh_xarray.lat.where(mesonh_xarray.lat <= 180, lambda lat: lat - 360),
        lon=mesonh_xarray.lon.where(mesonh_xarray.lon <= 180, lambda lon: lon - 360),
    )
    mesonh_xarray = mesonh_xarray.unify_chunks()

    dx, dy = guess_horizontal_spacing(mesonh_xarray, filename)

    # Add projection x and y coordinates to MesoNH
    proj_y_values = dy * (np.arange(0, mesonh_xarray.y.size) + 0.5)
    proj_x_values = dx * (np.arange(0, mesonh_xarray.x.size) + 0.5)

    mesonh_xarray["PROJY"] = ("y", proj_y_values)
    mesonh_xarray["PROJX"] = ("x", proj_x_values)

    if tracking_var.lower() == "dbz":
        mesonh_reflectivity = mesonh_calculate_reflectivity(mesonh_xarray)

        mesonh_xarray["DBZ"] = mesonh_reflectivity

        # add correct altitude based off of average height at each height index
        ht = mesonh_calculate_agl_z(mesonh_xarray)

        correct_alts = [np.mean(h.values) for h in ht]

        # Add altitude field for easier processing later
        mesonh_xarray["DBZ"] = mesonh_xarray["DBZ"].assign_coords(
            altitude=("z", correct_alts)
        )

    elif tracking_var.lower() == "tb":
        # Brightness temperature is only 2d so no heights needed
        mesonh_xarray["TB"] = (
            ["time", "y", "x"],
            mesonh_calculate_brightness_temp(mesonh_xarray),
        )
        mesonh_xarray["TB"].attrs["units"] = "K"

        # Adjust dask chunks
        mesonh_xarray["TB"] = mesonh_xarray["TB"].chunk(
            mesonh_xarray["LWup_TOA"].chunksizes
        )

    elif tracking_var.lower() == "wa":
        # Get updraft velocity at mass points (maybe?)
        mesonh_wa = mesonh_xarray.w

        mesonh_xarray["WA"] = mesonh_wa

        # Add correct altitude based off of average height at each height index
        ht = mesonh_calculate_agl_z(mesonh_xarray)

        correct_alts = [np.mean(h.values) for h in ht]

        # Add altitude field for easier processing later
        mesonh_xarray["WA"] = mesonh_xarray["WA"].assign_coords(
            altitude=("z", correct_alts)
        )

    elif tracking_var.lower() == "pr":
        # Brightness temperature is only 2d so no heights needed
        mesonh_xarray["PR"] = mesonh_xarray["pcp_rate"]

        mesonh_xarray["PR"].attrs["units"] = "mm/unit"

        # Adjust dask chunks
        mesonh_xarray["PR"] = mesonh_xarray["PR"].chunk(
            mesonh_xarray["LWup_TOA"].chunksizes
        )

    else:
        # If not any of the above, try using user inputed value
        try:
            var_values = mesonh_xarray[tracking_var]

            if len(var_values.shape) == 4:
                # add correct altitude based off of average height at each height index
                ht = mesonh_calculate_agl_z(mesonh_xarray)

                correct_alts = [np.mean(h.values) for h in ht]

                # Add altitude field for easier processing later
                mesonh_xarray[tracking_var] = mesonh_xarray[tracking_var].assign_coords(
                    altitude=("bottom_top", correct_alts)
                )

        except:
            raise Exception(
                f"!=====Invalid Tracking Variable. You Entered: {tracking_var}=====!"
            )

    return mesonh_xarray.unify_chunks()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jul 11 19:37:09 2024

@author: thahn
"""

# =============================================================================
# This file contains the functionality for detecting merging and splitting cells
# =============================================================================

import warnings
from copy import deepcopy

import numpy as np
import pandas as pd
import xarray as xr
from dask.array import isin
from tqdm import tqdm

from .irregularity.projection_calc_3d import (
    calc_3d_perim,
    calculate_volume,
    create_gridded_bounds,
)


def merge_split(
    analysis_object: dict,
    merge_analysis_type: str = "2d",
    **args: dict,
) -> tuple[pd.DataFrame, pd.DataFrame] | None:
    """


    Parameters
    ----------
    analysis_object : dict
        Standard CoCoCoMET-US Analysis_Object.
    merge_analysis_type : str, optional
        Type of merging and splitting to analyze. The default is "2d".
    **args : dict
        Throw away args.

    Raises
    ------
    ValueError
        If missing segmentation type needed for desired merging/splitting type.

    Returns
    -------
    merge_df : pandas.core.frame.DataFrame
        A pandas dataframe with these columns: frame, parent_cells, merged_cell.
    split_df : pandas.core.frame.DataFrame
        A pandas dataframe with these columns: frame, split_cell, child_cells.

    """

    if (
        merge_analysis_type.lower() == "3d"
        and analysis_object["US_segmentation_3d"] is not None
    ):

        merge_df, split_df = _merge_split_3d(analysis_object, **args)

    elif (
        merge_analysis_type.lower() == "2d"
        and (analysis_object["US_segmentation_2d"] is not None or analysis_object["US_segmentation_3d"] is not None)
    ):

        merge_df, split_df = _merge_split_2d(analysis_object, **args)

    else:
        raise ValueError("!=====Missing Desired Segmentation Input=====!")

    return merge_df, split_df


# Calculate nearest item in list to given pivot
def find_nearest(array, pivot):
    array = np.asarray(array)
    idx = (np.abs(array - pivot)).argmin()
    return idx


def _merge_split_2d(
    analysis_object: dict,
    variable: str,
    invert: bool = False,
    cell_footprint_height: float = 2,
    touching_threshold: float = 0.20,
    flood_background: float = 20,
    score_threshold: float = 0,
    score_weight_1: float = 1,
    score_weight_2: float = 1,
    radius_multiplyer: float = 0.1,
    overlap_threshold: float = 0.5,
    steps_forward_back: int = 2,
    height: int | None = None,
    **args: dict,
) -> tuple[pd.DataFrame, pd.DataFrame] | None:
    """


    Parameters
    ----------
    analysis_object : dict
        A  CoCoMET-US standard analysis object containing at least segmentation_xarray, US_tracks. and US_segmentation_2d or US_segmentation_3d, and segmentation_xarray.
    variable : str
        The variable which defines the background field.
    invert : bool, optional
        If we care about tracking lower values such as brightness temperature (i.e. we want to flood fill stuff less than a threshold), we need to invert the data so we can use the same algorithm. Will also invert background value. The default is False.
    cell_footprint_height : float, optional
        The height at which we want to find the newly calculated cell areas in kilometers. The default is 2km.
    touching_threshold : float, optional
        The percentage of cell border which must be shared by two cells. The default is 0.20.
    flood_background : float, optional
        The minimum background value for area calculation. The default is 20.
    score_threshold : float, optional
        The minimum value of the score function to be considered a part of the cell. The default is 0.
    score_weight_1 : float, optional
        The weighting of the relative strength of the cell. The default is 1.
    score_weight_2 : float, optional
        The weighting of the relative distance of the cell. The default is 1.
    radius_multiplyer : float, optional
        The multiplyer to increase the search radius by. Is equal to 1 + radius_multiplyer. The default is 0.1.
    overlap_threshold : float, optional
        The amount two cells need to overlap [0 to 1]. The default is 0.5.
    steps_forward_back : int, optional
        The number of steps to look forward or back to detect mergers/splitters. The default is 2.
    height:
        The height in km to calculate mergers and splitters at in 2D if variable if 3D. The default is None.
    **args : dict
        Throw away params.

    Raises
    ------
    Exception
        Exception if missing segmentation input.
        Exception if variable is 3D and height is None.

    Returns
    -------
    merged_df : pandas.core.frame.DataFrame
        A pandas dataframe with these columns: frame, parent_cells, merged_cell
    split_df : pandas.core.frame.DataFrame
        A pandas dataframe with these columns: frame, split_cell, child_cells
    """

    Tracks = analysis_object["US_tracks"]

    # If input variable field is 2D return None. Also, if DataArray, use those values for calculations. If Dataset, use tracking_var to get variable
    if type(analysis_object["segmentation_xarray"]) == xr.core.dataarray.DataArray:
        if len(analysis_object["segmentation_xarray"].shape) < 3:
            warnings.warn("!=====Input Variable has Too Low Dimensionality=====!")
            return None

        variable_field = analysis_object["segmentation_xarray"]

    else:
        if len(analysis_object["segmentation_xarray"][variable].shape) < 3:
            warnings.warn("!=====Input Variable has Too Low Dimensionality=====!")
            return None

        variable_field = analysis_object["segmentation_xarray"][variable]

    # If 3D segmentation is available, use that to calculate cell footprint, otherwise use 2D segmentation
    if analysis_object["US_segmentation_3d"] is not None:
        height_index = find_nearest(
            analysis_object["US_segmentation_3d"].altitude.values,
            cell_footprint_height * 1000,
        )

        footprint_data = analysis_object["US_segmentation_3d"].Feature_Segmentation[
            :, height_index
        ]

    elif analysis_object["US_segmentation_2d"] is not None:
        footprint_data = analysis_object["US_segmentation_2d"].Feature_Segmentation

        # if there is a 3D variable field, but only 2D segmentation, reduce the variable field dimensionality
        if (
            len(variable_field.shape) == 4 and variable_field.shape[1] != 1
        ):  # radar observations are 2D, but still have a z axis
            if height is None:
                raise Exception(
                    '2D merging and splitting on a 3D variable requires a "height" parameter. Either do 3D merging and splitting or set height'
                )

            try:  # TODO: reconcile this later, the models have attribute altitudes, but the observations have attribute altitude
                altitudes = analysis_object["segmentation_xarray"]["altitude"]
            except:
                altitudes = analysis_object["segmentation_xarray"]["altitudes"]

            vf_height_index = find_nearest(altitudes, height * 1000)
            variable_field = variable_field[:, vf_height_index]

    else:
        raise ValueError("!=====Missing Segmentation Input=====!")

    # If we care about tracking lower values such as brightness temperature (i.e. we want to flood fill stuff less than a threshold), we need to invert the data so we can use the same algorithm. Will also invert background value
    if invert:
        variable_field = -1 * variable_field
        flood_background = -1 * flood_background

    output_frame_list_merge = []
    output_init_cells_merge = []
    output_merged_cells = []

    output_frame_list_split = []
    output_init_cells_split = []
    output_split_cells = []

    frame_groups = Tracks.groupby("frame")

    for frame in tqdm(
        frame_groups,
        desc="=====Calculating Mergers and Splitters=====",
        total=frame_groups.ngroups,
    ):
        if frame[1].frame.min() >= (Tracks.frame.max() - (steps_forward_back - 1)):
            continue
        if frame[1].frame.min() <= (Tracks.frame.min() + (steps_forward_back - 1)):
            continue

        # Replace empty grid boxes with -1s to keep consistent with US
        final_mask = deepcopy(footprint_data[frame[1].frame.min()])
        final_mask.values[:] = -1

        for feature in frame[1].iterrows():
            feature_mask = deepcopy(footprint_data[frame[1].frame.min()])
            feature_mask.values[~isin(feature_mask.values, feature[1].feature_id)] = -1

            # Replace with cell id instead of feature id and merge with final_mask
            final_mask.values[feature_mask.values != -1] = feature[1].cell_id

            # If cell has no area, skip
            if np.all(feature_mask.values == -1):
                continue

        # Now find touching cells
        valid_cells = np.unique(final_mask.values)[1:]
        cell_data = final_mask.values

        ci = []
        ne = []
        tu = []
        for cell_id in valid_cells:
            num_of_edges = 0
            touching = []

            # Get indices of valid cells in mask
            # Loop over indices
            for nx, ny in zip(*np.where(cell_data == cell_id)):
                # Find if location is on edge of cell (i.e. any touching cells not)
                neighboring_cells = []

                for mx in range(nx - 1, nx + 2):
                    for my in range(ny - 1, ny + 2):
                        try:
                            t = cell_data[mx, my]
                        except IndexError:
                            t = -1

                        neighboring_cells.append(t)

                # Get all 9 adjecent data cells to the current data cell of interest
                neighboring_cells = np.array(neighboring_cells).reshape((3, 3))

                # If touching any data cells not belonging to the cell of interest, consider that cell an edge
                if np.sum(neighboring_cells != cell_id) != 0:
                    num_of_edges += 1

                temp = neighboring_cells[neighboring_cells != -1]
                temp = temp[temp != cell_id]
                touching.append(np.unique(temp))

            touching = np.concatenate(touching)

            ci.append(cell_id)
            ne.append(num_of_edges)
            tu.append(touching)

        cell_info_df = pd.DataFrame(
            data={"cell": ci, "Num_Edges": ne, "Touching_Edges": tu}
        )

        valid_touching_cell_sets = []

        # As long as one of the cells exceeds out threshold of touching, it will get added to tracked list, so no need to do anything more complex
        for cell in cell_info_df.iterrows():
            uu = np.unique(cell[1].Touching_Edges, return_counts=True)

            for touching_cellid, touching_edge_count in zip(uu[0], uu[1]):
                # If touching by over a certain percent threshold, add to valid touching set
                if (touching_edge_count / cell[1].Num_Edges) > touching_threshold:
                    touching_tuple = tuple(np.sort((cell[1].cell, touching_cellid)))
                    valid_touching_cell_sets.append(touching_tuple)

        # Save only unique tuples so tracking doesn"t repeat
        valid_touching_cell_sets = list(dict.fromkeys(valid_touching_cell_sets))

        valid_overlap_cell_sets = []

        for cell_set in valid_touching_cell_sets:
            # Cell 1 and 2 checks
            cell1_data = frame[1].query("cell_id==@cell_set[0]")
            cell2_data = frame[1].query("cell_id==@cell_set[1]")

            # If 3d, choose height, otherwise, use whole thing
            if (
                len(variable_field.shape) == 4
                and analysis_object["US_segmentation_3d"] is not None
            ):
                height_index = find_nearest(
                    analysis_object["US_segmentation_3d"].altitude.values,
                    cell_footprint_height * 1000,
                )

                reflectivity_data = deepcopy(
                    variable_field[cell1_data.frame.min()]
                ).values[height_index]

            else:
                reflectivity_data = deepcopy(
                    variable_field[cell1_data.frame.min()].squeeze()
                ).values

            # flood fill cell 1
            row_1 = int(np.round(cell1_data.south_north.values[0]))
            col_1 = int(np.round(cell1_data.west_east.values[0]))

            # Calculate the cell adjusted variable field, radius and then get the max distance of the search radius
            adj_Rmax_1 = np.round(reflectivity_data[row_1, col_1] - flood_background)
            cell_1_radius = np.ceil(
                cell_info_df.query("cell==@cell_set[0]").Num_Edges.values[0]
                / (2 * np.pi)
            )
            cell_1_radius = int(np.ceil((1 + radius_multiplyer) * cell_1_radius))
            max_dist_1 = np.sqrt(2) * cell_1_radius
            segmented_1 = np.zeros(reflectivity_data.shape)

            # Loop over search radius
            for mx in range(row_1 - cell_1_radius, row_1 + cell_1_radius + 1):
                for my in range(col_1 - cell_1_radius, col_1 + cell_1_radius + 1):
                    try:
                        # Calculate score function
                        R_adj = reflectivity_data[mx, my] - flood_background
                        dis = np.sqrt((my - col_1) ** 2 + (mx - row_1) ** 2)

                        # If the adjusted variable field value is not finite, skip it
                        if not np.isfinite(R_adj):
                            continue

                        # If the cell is the center of the search radius, will be a part of the cell
                        if dis == 0:
                            segmented_1[mx, my] = 1
                            continue

                        # If the maximum variable value or maximum distance is 0, just skip since every data cell would be counted
                        if adj_Rmax_1 == 0 or max_dist_1 == 0:
                            continue

                        # Calculate the score and see if it exceeds the threshold
                        score = score_weight_1 * (
                            R_adj / adj_Rmax_1
                        ) - score_weight_2 * (dis / max_dist_1)

                        if score > score_threshold:
                            segmented_1[mx, my] = 1

                    except:
                        continue

            segmented_1_area = np.sum(segmented_1)

            # flood fill cell 2
            row_2 = int(np.round(cell2_data.south_north.values[0]))
            col_2 = int(np.round(cell2_data.west_east.values[0]))

            # Calculate adjusted max variable, radius values and then max distance
            adj_Rmax_2 = np.round(reflectivity_data[row_2, col_2] - flood_background)
            cell_2_radius = np.ceil(
                cell_info_df.query("cell==@cell_set[1]").Num_Edges.values[0]
                / (2 * np.pi)
            )
            cell_2_radius = int(np.ceil(1.1 * cell_2_radius))
            max_dist_2 = np.sqrt(2) * cell_2_radius
            segmented_2 = np.zeros(reflectivity_data.shape)

            for mx in range(row_2 - cell_2_radius, row_2 + cell_2_radius + 1):
                for my in range(col_2 - cell_2_radius, col_2 + cell_2_radius + 1):
                    try:
                        # Repeat same process but for the second cell in the pair
                        R_adj = reflectivity_data[mx, my] - flood_background
                        dis = np.sqrt((my - col_2) ** 2 + (mx - row_2) ** 2)

                        if not np.isfinite(R_adj):
                            continue

                        if dis == 0:
                            segmented_2[mx, my] = 1
                            continue

                        if adj_Rmax_2 == 0 or max_dist_2 == 0:
                            continue

                        score = (R_adj / adj_Rmax_2) - (dis / max_dist_2)

                        if score > score_threshold:
                            segmented_2[mx, my] = 1

                    except:
                        continue

            segmented_2_area = np.sum(segmented_2)

            overlap_mask = np.logical_and(segmented_1, segmented_2)
            overlap_mask_area = np.sum(overlap_mask)

            if segmented_1_area == 0 or segmented_2_area == 0:
                continue

            max_ol_p = np.max(
                (
                    overlap_mask_area / segmented_1_area,
                    overlap_mask_area / segmented_2_area,
                )
            )

            if max_ol_p > overlap_threshold:
                valid_overlap_cell_sets.append(cell_set)

        # Check if all valid overlap cells still exist in the next frame(s), if no, whichever still exists merged
        for cell_set in valid_overlap_cell_sets:
            cell_1_status = [False] * steps_forward_back
            cell_2_status = [False] * steps_forward_back

            for step in range(1, steps_forward_back + 1):
                nf_id = frame[1].frame.min() + step
                next_frame = Tracks.query("frame==@nf_id")

                cell_1_status[step - 1] = (
                    next_frame.query("cell_id==@cell_set[0]").shape[0] != 0
                )
                cell_2_status[step - 1] = (
                    next_frame.query("cell_id==@cell_set[1]").shape[0] != 0
                )

            if np.all(cell_1_status) and not np.any(cell_2_status):
                output_frame_list_merge.append(
                    np.array([frame[1].frame.min(), frame[1].frame.min() + 1])
                )
                output_init_cells_merge.append(np.array((cell_set[0], cell_set[1])))
                output_merged_cells.append(cell_set[0])

            elif not np.any(cell_1_status) and np.all(cell_2_status):
                output_frame_list_merge.append(
                    np.array([frame[1].frame.min(), frame[1].frame.min() + 1])
                )
                output_init_cells_merge.append(np.array((cell_set[0], cell_set[1])))
                output_merged_cells.append(cell_set[1])

        # Check if all valid overlap cells still exist in the previous frame(s), if no, whichever didn"t exist is the split
        for cell_set in valid_overlap_cell_sets:
            cell_1_status = [False] * steps_forward_back
            cell_2_status = [False] * steps_forward_back

            for step in range(1, steps_forward_back + 1):
                nf_id = frame[1].frame.min() - step
                next_frame = Tracks.query("frame==@nf_id")

                cell_1_status[step - 1] = (
                    next_frame.query("cell_id==@cell_set[0]").shape[0] != 0
                )
                cell_2_status[step - 1] = (
                    next_frame.query("cell_id==@cell_set[1]").shape[0] != 0
                )

            if np.all(cell_1_status) and not np.any(cell_2_status):
                output_frame_list_split.append(
                    np.array([frame[1].frame.min() - 1, frame[1].frame.min()])
                )
                output_init_cells_split.append(cell_set[0])
                output_split_cells.append((cell_set[0], cell_set[1]))
            elif not np.any(cell_1_status) and np.all(cell_2_status):
                output_frame_list_split.append(
                    np.array([frame[1].frame.min() - 1, frame[1].frame.min()])
                )
                output_init_cells_split.append(cell_set[1])
                output_split_cells.append((cell_set[0], cell_set[1]))

    merged_df = pd.DataFrame(
        data={
            "frame": output_frame_list_merge,
            "parent_cells": output_init_cells_merge,
            "merged_cell": output_merged_cells,
        }
    )
    split_df = pd.DataFrame(
        data={
            "frame": output_frame_list_split,
            "split_cell": output_init_cells_split,
            "child_cells": output_split_cells,
        }
    )

    return (merged_df, split_df)


def _merge_split_3d(
    analysis_object: dict,
    variable: str | None = None,
    invert: bool = False,
    touching_threshold: float = 0.2,
    flood_background: float = 20,
    radius_multiplyer: float = 0.1,
    score_threshold: float = 0,
    score_weight_1: float = 1,
    score_weight_2: float = 1,
    overlap_threshold: float = 0.5,
    **args: dict,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Parameters
    ----------
    analysis_dict : dict
        A CoCoMET-US standard analysis object containing at least US_segmentation_3d, US_tracks, and segmentation_xarray.
    variable : str
        The variable which defines the background field.
    invert : bool, optional
        If we care about tracking lower values such as brightness temperature (i.e. we want to flood fill stuff less than a threshold), we need to invert the data so we can use the same algorithm. Will also invert background value. The default is False.    For fields where we care about tracking minimums
    touching_threshold : float, optional
        The percentage of cell border which must be shared by two cells. The default is 0.20.
    flood_background : float, optional
        The minimum background value for area calculation. The default is 20.
    score_threshold : float, optional
        The minimum value of the score function to be considered a part of the cell. The default is 0.
    score_weight_1 : float, optional
        The weighting of the relative strength of the cell. The default is 1.
    score_weight_2 : float, optional
        The weighting of the relative distance of the cell. The default is 1.
    radius_multiplyer : float, optional
        The multiplyer to increase the search radius by. Is equal to 1 + radius_multiplyer. The default is 0.1.
    overlap_threshold : float, optional
        The amount two cells need to overlap [0 to 1]. The default is 0.5.
    **args : dict
        Throw away params.

    Raises
    ------
    Exception
        Exception if missing 3D segmentation input.

    Returns
    -------
    merged_df : pandas.DataFrame
        Dataframe with the following columns:
        frame: Tuple of frames before and after the merger
        parent_cells: Tuple of two parent cell ids. The first entry of the tuple is the cell that survives the merger
        merged_cell: the cell id that survived the merger
    split_df : pandas.DataFrame
        Dataframe with the following columns:
        frame: Tuple of frames before and after the split
        split_cell: cell id of the original cell
        child_cells: tuple of cell ids of the 2 child cells. The first entry of the tuple is the original cell, and seconds is the cell that was created from the split.

    assumptions (will be expanded later): 3d input
    """

    Tracks = analysis_object["US_tracks"]

    # If input variable field is 2D return None. Also, if DataArray, use those values for calculations. If Dataset, use tracking_var to get variable
    if type(analysis_object["segmentation_xarray"]) == xr.core.dataarray.DataArray:
        if len(analysis_object["segmentation_xarray"].shape) < 3:
            warnings.warn("!=====Input Variable has Too Low Dimensionality=====!")
            return None

        variable_field = analysis_object["segmentation_xarray"]

    else:
        if len(analysis_object["segmentation_xarray"][variable].shape) < 3:
            warnings.warn("!=====Input Variable has Too Low Dimensionality=====!")
            return None

        variable_field = analysis_object["segmentation_xarray"][variable]

    # If we care about tracking lower values such as brightness temperature (i.e. we want to flood fill stuff less than a threshold), we need to invert the data so we can use the same algorithm. Will also invert background value
    if invert:
        variable_field = -1 * variable_field
        flood_background = -1 * flood_background

    footprint_data = analysis_object["US_segmentation_3d"].Feature_Segmentation

    # inner_Tracks = Tracks.loc[(Tracks['frame'] != 0) & (Tracks['frame'] != max(Tracks['frame'].tolist()))] #remove all rows for the first and last frame
    cells = Tracks["cell_id"]
    frames = Tracks["frame"]
    final_frame_indices = frames[
        frames == max(frames.tolist())
    ].index.tolist()  # indeces of Tracks with the last frame
    first_frame_indices = frames[
        frames == 0
    ].index.tolist()  # indeces with the first frame
    temp = (
        []
    )  # list of Tracks indices where a cell appeared for the last time (these might be where mergers occured)
    for cell_id in np.unique(cells.tolist()):
        if cell_id == -1:
            continue
        temp.append(cells.where(cells == cell_id).last_valid_index())
    last_appearances = [
        i for i in temp if i not in final_frame_indices
    ]  # ignore any last appearances on the final frame

    temp = []
    for cell_id in np.unique(cells.tolist()):
        if cell_id == -1:
            continue
        temp.append(cells.where(cells == cell_id).first_valid_index())
    first_appearences = [
        i for i in temp if i not in first_frame_indices
    ]  # ignore any first appearances on the first frame

    output_frame_list_merge = []
    output_init_cells_merge = []
    output_merged_cells = []

    output_frame_list_split = []
    output_init_cells_split = []
    output_split_cells = []

    fi = []  # feature ids
    sa = []  # surface areas
    sad = []  # surface area dictionaries (from calc_perim_3d function)
    merge_or_split_list = []

    # use one loop for both the merging and the splitting, but there are a few changes that need to be made, so this merge_split_bool is defined
    # True indicates we are doing mergers and false indicates splits
    for i, index_list in enumerate([last_appearances, first_appearences]):
        if i == 0:
            merge_split_bool = True
            desc = "=====Calculating Merger Edges====="
        else:
            merge_split_bool = False
            desc = "=====Calculating Split Edges====="

        for i in tqdm(
            index_list, desc=desc, total=len(index_list)
        ):  # loop over every feature's final frame (1 cell per loop)
            row = Tracks.iloc[i]
            feature_id = row["feature_id"]
            frame = row["frame"]
            cell_id = row["feature_id"]
            # position = (row['projection_x'], row['projection_y'], row['altitude'])
            # velocity = need to think about how to get this. maybe import my velocity calculation or find if CoCoMET calculates it somewhere

            # this is a dictionary with keys equal to what the cell is touching (-1 for empty space and feature_id for other cells)
            # values are the total surface area that is touching that thing
            surface_dict = calc_3d_perim(footprint_data, frame, feature_id)
            if surface_dict == -1:
                continue
            total_surface_area = sum(surface_dict.values())

            fi.append(feature_id)
            sa.append(total_surface_area)
            sad.append(surface_dict)

            if merge_split_bool:
                merge_or_split_list.append("merge")
            else:
                merge_or_split_list.append("split")

    cell_info_df = pd.DataFrame(
        data={
            "feature_id": fi,
            "Surface_Area": sa,
            "Surface_Area_Dict": sad,
            "Merge_or_Split": merge_or_split_list,
        }
    )

    cell_info_copy = cell_info_df.copy()  # iterate over this to avoid issues

    valid_touching_cell_sets = []
    # As long as one of the cells exceeds out threshold of touching, it will get added to tracked list, so no need to do anything more complex
    for i, cell in tqdm(
        cell_info_copy.iterrows(),
        desc="=====Filtering By Touching %=====",
        total=len(cell_info_copy),
    ):  # iterates over rows of dataframe and returns (index, series) pairs
        # cell[1] is just the row of the dataframe

        total_surface = cell.Surface_Area

        for (
            touching_featureid,
            touching_area,
        ) in cell.Surface_Area_Dict.items():  # loop through surface area dictionary
            changing_cell = Tracks[Tracks["feature_id"] == cell.feature_id][
                "cell_id"
            ].values[0]

            if touching_featureid == -1:
                continue

            constant_cell = Tracks[Tracks["feature_id"] == touching_featureid][
                "cell_id"
            ].values[
                0
            ]  # define the two cell ids involved in the potential merge/split

            if constant_cell == -1 or changing_cell == -1:
                continue

            frame = Tracks[Tracks["feature_id"] == touching_featureid]["frame"].values[
                0
            ]

            touching_cell_surface_dict = calc_3d_perim(
                footprint_data, frame, touching_featureid
            )
            if touching_cell_surface_dict == -1:
                continue
            touching_cell_surface_area = sum(touching_cell_surface_dict.values())

            # If touching by over a certain percent threshold, add to valid touching set
            if (touching_area / total_surface) > touching_threshold or (
                touching_area / touching_cell_surface_area
            ) > touching_threshold:
                # this if statement is to avoid the case where 2 cells are touching but the both disappear on the next frame, which would not be a merge
                if (
                    cell.Merge_or_Split == "merge"
                    and Tracks.iloc[
                        cells.where(cells == changing_cell).last_valid_index()
                    ]["frame"]
                    == Tracks.iloc[
                        cells.where(cells == constant_cell).last_valid_index()
                    ]["frame"]
                ):
                    continue
                # same here if 2 cells suddenly appear on the same frame then that isn't a split
                if (
                    cell.Merge_or_Split == "split"
                    and Tracks.iloc[
                        cells.where(cells == changing_cell).first_valid_index()
                    ]["frame"]
                    == Tracks.iloc[
                        cells.where(cells == constant_cell).first_valid_index()
                    ]["frame"]
                ):
                    continue

                # add the touching cell to the cell info df
                # we only need the number of edges, so the other values are just set to -1

                cell_info_df.loc[len(cell_info_df.index)] = [
                    touching_featureid,
                    touching_cell_surface_area,
                    touching_cell_surface_dict,
                    -1,
                ]

                # (constant cell, changing cell, merge/split)
                touching_tuple = (
                    touching_featureid,
                    cell.feature_id,
                    cell.Merge_or_Split,
                )
                valid_touching_cell_sets.append(
                    touching_tuple
                )  # tuple of touching cells, so (1, 2) means feature_id 1 and 2 are touching

    valid_overlap_cell_sets = []
    for cell_set in tqdm(
        valid_touching_cell_sets,
        desc="=====Calculating Overlap=====",
        total=len(valid_touching_cell_sets),
    ):
        # Cell 1 and 2 checks
        cell1_data = Tracks[
            Tracks["feature_id"] == cell_set[0]
        ]  # find rows in frame[1] where cell_id == cell_set[0/1]. cell_set is a pair of touching cells
        cell2_data = Tracks[Tracks["feature_id"] == cell_set[1]]

        # Check if data is 3D
        if (
            len(variable_field.shape) == 4
            and analysis_object["US_segmentation_3d"] is not None
        ):
            variable_data = deepcopy(variable_field[cell1_data.frame.values[0]]).values

        else:
            raise ValueError("!=====Variable Field has Wrong Dimension=====!")

        # flood fill cell 1

        # doesnt this section find the top left corner of a cell when it should be finding the center
        # No, south_north only contains one value which is the value of the feature, so this is actually correct.
        row_1 = int(np.round(cell1_data.up_down.values[0]))
        col_1 = int(np.round(cell1_data.south_north.values[0]))
        depth_1 = int(np.round(cell1_data.west_east.values[0]))

        # Calculate the cell adjusted variable field, radius and then get the max distance of the search radius
        adj_Rmax_1 = np.round(
            variable_data[row_1, col_1, depth_1] - flood_background
        )  # variable at the feature minus the background. Still need to implement the feature location
        cell_1_radius = np.sqrt(
            cell_info_df[cell_info_df["feature_id"] == cell_set[0]].Surface_Area.values[
                0
            ]
            / (4 * np.pi)
        )
        cell_1_radius = (1 + radius_multiplyer) * cell_1_radius
        max_dist_1 = np.sqrt(3) * cell_1_radius
        segmented_1_points = []
        # Loop over search radius

        center_proj_1 = (
            cell1_data.altitude.values[0],
            cell1_data.projection_y.values[0],
            cell1_data.projection_x.values[0],
        )
        zbounds_1, ybounds_1, xbounds_1 = create_gridded_bounds(
            footprint_data, center_proj_1, cell_1_radius
        )
        for mx in range(xbounds_1[0], xbounds_1[1]):
            for my in range(ybounds_1[0], ybounds_1[1]):
                for mz in range(zbounds_1[0], zbounds_1[1]):
                    try:
                        # Calculate score function
                        R_adj = variable_data[mz, my, mx] - flood_background
                        dis = np.sqrt(
                            (my - col_1) ** 2 + (mz - row_1) ** 2 + (mx - depth_1) ** 2
                        )

                        # If the adjusted variable field value is not finite, skip it (i guess reflectivity could be NaN)
                        if not np.isfinite(R_adj):
                            continue

                        # If the cell is the center of the search radius, will be a part of the cell
                        if dis == 0:
                            segmented_1_points.append((mz, my, mx))
                            continue

                        # If the maximum variable value or maximum distance is 0, just skip since every data cell would be counted
                        if adj_Rmax_1 == 0 or max_dist_1 == 0:
                            continue  # need to check for divide by zero errors any time a division happens (below)

                        # Calculate the score and see if it exceeds the threshold
                        score = score_weight_1 * (
                            R_adj / adj_Rmax_1
                        ) - score_weight_2 * (dis / max_dist_1)

                        if score > score_threshold:
                            segmented_1_points.append((mz, my, mx))

                    except:
                        continue

        segmented_1_area = calculate_volume(footprint_data, segmented_1_points)

        # flood fill cell 2
        row_2 = int(np.round(cell2_data.up_down.values[0]))
        col_2 = int(np.round(cell2_data.south_north.values[0]))
        depth_2 = int(np.round(cell2_data.west_east.values[0]))

        # Calculate adjusted max variable, radius values and then max distance
        adj_Rmax_2 = np.round(variable_data[row_2, col_2, depth_2] - flood_background)
        cell_2_radius = np.ceil(
            np.sqrt(
                cell_info_df[
                    cell_info_df["feature_id"] == cell_set[1]
                ].Surface_Area.values[0]
                / (4 * np.pi)
            )
        )
        cell_2_radius = int(np.ceil(1.1 * cell_2_radius))
        max_dist_2 = np.sqrt(3) * cell_2_radius
        segmented_2_points = []

        center_proj_2 = (
            cell1_data.altitude.values[0],
            cell1_data.projection_y.values[0],
            cell1_data.projection_x.values[0],
        )
        zbounds_2, ybounds_2, xbounds_2 = create_gridded_bounds(
            footprint_data, center_proj_2, cell_2_radius
        )
        for mx in range(xbounds_2[0], xbounds_2[1]):
            for my in range(ybounds_2[0], ybounds_2[1]):
                for mz in range(zbounds_2[0], zbounds_2[1]):
                    try:
                        # Repeat same process but for the second cell in the pair
                        R_adj = variable_data[mz, my, mx] - flood_background
                        dis = np.sqrt(
                            (my - col_2) ** 2 + (mz - row_2) ** 2 + (mx - depth_2) ** 2
                        )

                        if not np.isfinite(R_adj):
                            continue

                        if dis == 0:
                            segmented_2_points.append((mz, my, mx))
                            continue

                        if adj_Rmax_2 == 0 or max_dist_2 == 0:
                            continue

                        score = score_weight_1 * (
                            R_adj / adj_Rmax_2
                        ) - score_weight_2 * (
                            dis / max_dist_2
                        )  # im assuming score weights should go here aswell

                        if score > score_threshold:
                            segmented_2_points.append((mz, my, mx))

                    except:
                        continue

        segmented_2_area = calculate_volume(footprint_data, segmented_2_points)

        overlap_points = list(set(segmented_1_points) & set(segmented_2_points))
        overlap_mask_area = calculate_volume(footprint_data, overlap_points)

        if segmented_1_area == 0 or segmented_2_area == 0:
            continue

        max_ol_p = np.max(
            (
                overlap_mask_area / segmented_1_area,
                overlap_mask_area / segmented_2_area,
            )
        )

        if max_ol_p > overlap_threshold:
            valid_overlap_cell_sets.append(cell_set)

    # Loop through all remaining cells and add them to the ouput
    for cell_set in tqdm(
        valid_overlap_cell_sets,
        desc="=====Creating Output Dataframe=====",
        total=len(valid_overlap_cell_sets),
    ):
        frame = Tracks[Tracks["feature_id"] == cell_set[0]]["frame"].values[0]
        cell_id1 = Tracks[Tracks["feature_id"] == cell_set[0]]["cell_id"].values[0]
        cell_id2 = Tracks[Tracks["feature_id"] == cell_set[1]]["cell_id"].values[0]
        if cell_set[2] == "merge":
            output_frame_list_merge.append((frame, frame + 1))
            output_init_cells_merge.append((cell_id1, cell_id2))
            output_merged_cells.append(cell_id1)
        else:
            output_frame_list_split.append((frame - 1, frame))
            output_init_cells_split.append(cell_id1)
            output_split_cells.append((cell_id1, cell_id2))

    merged_df = pd.DataFrame(
        data={
            "frame": output_frame_list_merge,
            "parent_cells": output_init_cells_merge,
            "merged_cell": output_merged_cells,
        }
    )
    split_df = pd.DataFrame(
        data={
            "frame": output_frame_list_split,
            "split_cell": output_init_cells_split,
            "child_cells": output_split_cells,
        }
    )

    return (merged_df, split_df)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul 22 11:03:37 2024

@author: thahn
"""

# =============================================================================
# This defines the analysis object class
# =============================================================================


class Analysis_Object:
    def __init__(
        self,
        tracking_xarray,
        segmentation_xarray,
        US_features,
        US_tracks,
        US_segmentation_2d,
        US_segmentation_3d,
    ):
        self.tracking_xarray = tracking_xarray
        self.segmentation_xarray = segmentation_xarray
        self.US_features = US_features
        self.US_tracks = US_tracks
        self.US_segmentation_2d = US_segmentation_2d
        self.US_segmentation_3d = US_segmentation_3d

    def return_analysis_dictionary(self) -> dict:
        return {
            "tracking_xarray": self.tracking_xarray,
            "segmentation_xarray": self.segmentation_xarray,
            "US_features": self.US_features,
            "US_tracks": self.US_tracks,
            "US_segmentation_2d": self.US_segmentation_2d,
            "US_segmentation_3d": self.US_segmentation_3d,
        }
